import 'package:oobium_datastore_gen/src/generators/rest_feature_adapter.dart';
import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';
import 'package:xstring/xstring.dart';

String genRest(Schema schema) => RestFeatureGenerator(schema).compile();

class RestFeatureGenerator {

  final Schema schema;
  Iterable<Model> get models => schema.models;
  RestFeatureGenerator(this.schema);

  String compileCreator() =>
'''
  _RestFeature(
    ${models.map((m) => '${m.restAdapterVarName}: ${m.restAdapterVarName} ?? ${m.restAdapterName}(),').join()}
  )
''';

  String compile() =>
'''
${schema.partOfDeclaration}

class _RestFeature extends BufferedStateFeature {
  late final ${schema.dsName} _modelContext;
  ${models.map((m) => 'final ${m.restAdapterName} ${m.restAdapterVarName};').join()}
  _RestFeature({
    ${models.map((m) => 'required this.${m.restAdapterVarName},').join()}
  }) {
    ${models.map((m) => '${m.restAdapterVarName}._feature = this;').join()}
  }
  
  @override
  Future<void> onClose() {
    _expire();
    return super.onClose();
  }
  
  @override
  Future<void> onDestroy() {
    _expire();
    return super.onDestroy();
  }
  
  @override
  void onEvents(StateEvents events) {
    ${['created', 'updated', 'removed'].map((type) => '''
      for(final event in events.${type.varName}(this)) {${models.where((m) => (type != 'removed' || m.isNotSingleton)).map((m) => '''
        if(event is StateEvent<${m.snapName}>) {
          _log.debug(() => '$type-${m.name}');
          q + (ex) => ${m.restAdapterVarName}._$type(ex, ${(type == 'updated') ? 'event.oldValue, event.newValue' : 'event.value'});
        }''').join(' else ')}
        else {
          _log.warn(() => '$type-UnknownType: \$event');
        }
      }
    ''').join()}
  }
  
  void _expire() {
    ${models.where((m) => m.canFetch).map((m) => '${m.restAdapterVarName}.expire();').join()}
  }
}

abstract class _RestAdapter {
  late final _RestFeature _feature;
  ${schema.dsName} get _modelContext => _feature._modelContext;
}

${models.map((m) => RestAdapterGenerator(m).compile()).join('\n')}
''';
}

extension RestModelX on Model {
  String get restAdapterName => '${name}RestAdapter';
  String get restAdapterVarName => '${varName}RestAdapter';
}
