import 'package:oobium_datastore_gen/src/feature.dart';
import 'package:oobium_datastore_gen/src/generators/trigger_feature_adapter.dart';
import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';

String genTrigger(Schema schema) => TriggerFeatureGenerator(schema).compile();

class TriggerFeatureGenerator {

  final Schema schema;
  TriggerFeatureGenerator(this.schema);

  String get dsName => schema.dsName;

  bool get hasDependents => models.isNotEmpty;
  Iterable<Model> get models => schema.models;

  String compileCreator() =>
'''
  _TriggerFeature(
    ${models.map((m) => '${m.triggersVarName}: ${m.triggersVarName},').join()}
  )
''';

  String compile() =>
schema.has<TriggerFeature>() ?
'''
${schema.partOfDeclaration}

class _TriggerFeature extends StateFeature {
  late final ${schema.dsName} _modelContext;
  ${models.map((m) => 'final ${m.triggersName}? ${m.triggersVarName};').join()}
  _TriggerFeature({
    ${models.map((m) => 'this.${m.triggersVarName},').join()}
  });

  @override
  void onEvent(StateEvent e) {
    if(e.isCreate) _onCreate(e);
    else if(e.isUpdate) _onUpdate(e);
    else if(e.isRemove) _onRemove(e);
  }
  
  void _onCreate(StateEvent e) {${models.map((m) => '''
    if(e is StateEvent<${m.snapName}>) {
      _log.trace(() => 'created-${m.name}');
      ${m.triggersVarName}?.onCreate?.call(_modelContext, e.value);
    }''').join(' else ')}
  }
  
  void _onUpdate(StateEvent e) {${models.map((m) => '''
    if(e is StateEvent<${m.snapName}>) {
      _log.trace(() => 'updated-${m.name}');
      ${m.triggersVarName}?.onUpdate?.call(_modelContext, e.oldValue, e.newValue);
    }''').join(' else ')}
  }
  
  void _onRemove(StateEvent e) {${models.map((m) => '''
    if(e is StateEvent<${m.snapName}>) {
      _log.trace(() => 'removed-${m.name}');
      ${remove(m)}
    }''').join(' else ')}
  }
}

${models.map((m) => TriggerAdapterGenerator(m).compile()).join('\n')}
''' : '';

  String remove(Model m) => [
    ...m.dependentsFields.where((f) => f.model.isNotSingleton).map((f) =>
        f.isHasMany
            ? 'e.source.remove(${f.model.keyType}(e.value.${f.snapName}));'
            : 'for(final ${f.model.varName} in getAll<${f.model.snapName}>((s) => s.${f.snapName} == e.value.id)) {'
              '  e.source.remove(${f.model.varName});'
              '}'
        ),
    if(m.isNotSingleton)
      '${m.triggersVarName}?.onRemove?.call(_modelContext, e.value);'
  ].join();
}

extension TriggerModelX on Model {
  String get triggersName => '${name}Triggers';
  String get triggersVarName => '${varName}Triggers';
}
